#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# ~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~ #
# Standard library imports

# Third party imports
import pysam

# Local imports

# ~~~~~~~~~~~~~~CLASS READ~~~~~~~~~~~~~~ #
class Read(object):
    """
    Represent a read from the original fastq file associated with alignments found
    by minimap2
    """

    # ~~~~~~~~~~~~~~MAGIC METHODS~~~~~~~~~~~~~~ #
    def __init__(self):
        self.alignment_list = []

    def __repr__(self):
        m = ""
        for r in self.alignment_list:
            m += "\t\t{}\n".format(r)
        return m

    # ~~~~~~~~~~~~~~PROPERTY METHODS~~~~~~~~~~~~~~ #
    @property
    def n_alignment(self):
        return len(self.alignment_list)

    # ~~~~~~~~~~~~~~PUBLIC METHODS~~~~~~~~~~~~~~ #
    def get_best_alignment(self, primary_score="alignment_score"):
        primary_alignment_idx = self._get_primary_idx(primary_score=primary_score)
        return self.alignment_list[primary_alignment_idx]

    def get_secondary_alignments_list(self, primary_score="alignment_score"):
        primary_alignment_idx = self._get_primary_idx(primary_score=primary_score)
        return [
            alignment
            for i, alignment in enumerate(self.alignment_list)
            if i != primary_alignment_idx
        ]

    def add_pysam_alignment(self, pysam_aligned_segment, read_idx, **kwargs):
        self.alignment_list.append(Alignment(pysam_aligned_segment, read_idx))

    def add_alignment(self, alignment, **kwargs):
        self.alignment_list.append(alignment)

    # ~~~~~~~~~~~~~~PRIVATE METHODS~~~~~~~~~~~~~~ #
    def _get_primary_idx(self, primary_score="alignment_score"):
        idx = 0
        if primary_score == "alignment_score":
            best_align_score = -1
            for i, alignment in enumerate(self.alignment_list):
                if alignment.align_score > best_align_score:
                    best_align_score = alignment.align_score
                    idx = i
        elif primary_score == "primary":
            for i, alignment in enumerate(self.alignment_list):
                if not alignment.secondary:
                    idx = i
        elif primary_score == "alignment_length":
            best_align_len = -1
            for i, alignment in enumerate(self.alignment_list):
                if alignment.align_len > best_align_len:
                    best_align_len = alignment.align_len
                    idx = i
        return idx


# ~~~~~~~~~~~~~~CLASS HIT~~~~~~~~~~~~~~ #
class Alignment:
    """
    Helper class to extract relevant fields from the a pysam alignedSegment object
    """

    # ~~~~~~~~~~~~~~MAGIC METHODS~~~~~~~~~~~~~~ #
    def __init__(self, pysam_aligned_segment, read_idx):
        """"""
        self.qname = pysam_aligned_segment.query_name
        self.rname = pysam_aligned_segment.reference_name
        qlen = int(pysam_aligned_segment.query_length)
        if qlen == 0:
            self.qlen = int(pysam_aligned_segment.infer_query_length())
        else:
            self.qlen = qlen
        self.align_len = int(pysam_aligned_segment.query_alignment_length)
        self.align_score = int(pysam_aligned_segment.get_tag("AS"))
        self.secondary = (
            pysam_aligned_segment.is_secondary or pysam_aligned_segment.is_supplementary
        )
        self.read_idx = read_idx

    def __repr__(self):
        return "Query:{} | Reference:{} | Query len:{} | Alignment len:{} | Align Score:{} | Secondary:{}".format(
            self.qname,
            self.rname,
            self.qlen,
            self.align_len,
            self.align_score,
            self.secondary,
        )

    # ~~~~~~~~~~~~~~PROPERTY METHODS~~~~~~~~~~~~~~ #
    @property
    def query_fraction_aligned(self):
        if self.qlen == 0:
            return 0
        else:
            return self.align_len / self.qlen



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~ #

# Standard library imports
import argparse

# Local imports
import NanoCount as pkg
from NanoCount.NanoCount import NanoCount as nc
from NanoCount.common import *

# ~~~~~~~~~~~~~~MAIN PARSER ENTRY POINT~~~~~~~~~~~~~~ #


def main(args=None):

    # Define parser
    parser = argparse.ArgumentParser(description=pkg.__description__)
    parser.add_argument("--version", action="version", version="{} v{}".format(pkg.__name__, pkg.__version__))

    parser_io = parser.add_argument_group("Input/Output options")
    arg_from_docstr(parser=parser_io, func=nc, arg_name="alignment_file", short_name="i")
    arg_from_docstr(parser=parser_io, func=nc, arg_name="count_file", short_name="o")
    arg_from_docstr(parser=parser_io, func=nc, arg_name="filter_bam_out", short_name="b")

    parser_ms = parser.add_argument_group("Misc options")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="min_alignment_length", short_name="l")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="min_query_fraction_aligned", short_name="f")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="sec_scoring_value", short_name="s")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="sec_scoring_threshold", short_name="t")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="convergence_target", short_name="c")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="max_em_rounds", short_name="e")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="extra_tx_info", short_name="x")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="primary_score", short_name="p")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="keep_suplementary", short_name="a")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="max_dist_3_prime", short_name="d")
    arg_from_docstr(parser=parser_ms, func=nc, arg_name="max_dist_5_prime", short_name="u")

    parser_vb = parser.add_argument_group("Verbosity options")
    arg_from_docstr(parser=parser_vb, func=nc, arg_name="verbose", short_name="v")
    arg_from_docstr(parser=parser_vb, func=nc, arg_name="quiet", short_name="q")

    # Parse args and run main function
    args = parser.parse_args()
    nanocount = nc(**vars(args))


# execute only if run as a script
if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# ~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~ #
# Standard library imports
import sys
import os
from collections import *
import logging
import datetime
import inspect

# Third party imports
import colorlog
import NanoCount as pkg

# ~~~~~~~~~~~~~~FUNCTIONS~~~~~~~~~~~~~~ #


def opt_summary(local_opt):
    """Simplifiy option dict creation"""
    d = OrderedDict()
    d["Package name"] = pkg.__name__
    d["Package version"] = pkg.__version__
    d["Timestamp"] = str(datetime.datetime.now())
    for i, j in local_opt.items():
        if not i == "self":
            d[i] = j
    return d


def get_logger(name=None, verbose=False, quiet=False):
    """Multilevel colored log using colorlog"""

    # Define conditional color formatter
    formatter = colorlog.LevelFormatter(
        fmt={
            'DEBUG': '%(log_color)s\t[DEBUG]: %(msg)s',
            'INFO': '%(log_color)s\t%(msg)s',
            'WARNING': '%(log_color)s## %(msg)s ##',
            'ERROR': '%(log_color)sERROR: %(msg)s',
            'CRITICAL': '%(log_color)sCRITICAL: %(msg)s',
        },
        log_colors={
            'DEBUG': 'white',
            'INFO': 'green',
            'WARNING': 'bold_blue',
            'ERROR': 'bold_red',
            'CRITICAL': 'bold_purple',
        },
        reset=True,
    )

    # Define logger with custom formatter
    logging.basicConfig(format='%(message)s')
    logging.getLogger().handlers[0].setFormatter(formatter)
    log = logging.getLogger(name)

    # Define logging level depending on verbosity
    if verbose:
        log.setLevel(logging.DEBUG)
    elif quiet:
        log.setLevel(logging.WARNING)
    else:
        log.setLevel(logging.INFO)

    return log


def log_dict(d, logger, header="", indent="\t", level=1):
    """ log a multilevel dict """
    if header:
        logger(header)
    if isinstance(d, Counter):
        for i, j in d.most_common():
            logger("{}{}: {:,}".format(indent * level, i, j))
    else:
        for i, j in d.items():
            if isinstance(j, dict):
                logger("{}{}".format(indent * level, i, j))
                log_dict(j, logger, level=level + 1)
            else:
                logger("{}{}: {}".format(indent * level, i, j))


def log_list(l, logger, header="", indent="\t"):
    """ log a list """
    if header:
        logger(header)
    for i in l:
        logger("{}*{}".format(indent, i))


def doc_func(func):
    """Parse the function description string"""

    if inspect.isclass(func):
        func = func.__init__

    docstr_list = []
    for l in inspect.getdoc(func).split("\n"):
        l = l.strip()
        if l:
            if l.startswith("*"):
                break
            else:
                docstr_list.append(l)

    return " ".join(docstr_list)


def make_arg_dict(func):
    """Parse the arguments default value, type and doc"""

    # Init method for classes
    if inspect.isclass(func):
        func = func.__init__

    if inspect.isfunction(func) or inspect.ismethod(func):
        # Parse arguments default values and annotations
        d = OrderedDict()
        for name, p in inspect.signature(func).parameters.items():
            if not p.name in [
                "self",
                "cls",
            ]:  # Object stuff. Does not make sense to include in doc
                d[name] = OrderedDict()
                if not name in [
                    "kwargs",
                    "args",
                ]:  # Include but skip default required and type
                    # Get Annotation
                    if p.annotation != inspect._empty:
                        d[name]["type"] = p.annotation
                    # Get default value if available
                    if p.default == inspect._empty:
                        d[name]["required"] = True
                    else:
                        d[name]["default"] = p.default

        # Parse the docstring in a dict
        docstr_dict = OrderedDict()
        lab = None
        for l in inspect.getdoc(func).split("\n"):
            l = l.strip()
            if l:
                if l.startswith("*"):
                    lab = l[1:].strip()
                    docstr_dict[lab] = []
                elif lab:
                    docstr_dict[lab].append(l)

        # Concatenate and copy doc in main dict
        for name in d.keys():
            if name in docstr_dict:
                d[name]["help"] = " ".join(docstr_dict[name])
        return d


def arg_from_docstr(parser, func, arg_name, short_name=None):
    """Get options corresponding to argument name from docstring and deal with special cases"""

    if short_name:
        arg_names = ["-{}".format(short_name), "--{}".format(arg_name)]
    else:
        arg_names = ["--{}".format(arg_name)]

    arg_dict = make_arg_dict(func)[arg_name]

    if "help" in arg_dict:
        if "default" in arg_dict:
            if arg_dict["default"] == "" or arg_dict["default"] == []:
                arg_dict["help"] += " (default: None)"
            else:
                arg_dict["help"] += " (default: %(default)s)"
        else:
            arg_dict["help"] += " (required)"

        if "type" in arg_dict:
            if arg_dict["type"] == bool:
                arg_dict["help"] += " [boolean]"
            else:
                arg_dict["help"] += " [%(type)s]"

    # Special case for boolean args
    if arg_dict["type"] == bool:
        del arg_dict["type"]
        if arg_dict["default"] == False:
            arg_dict["action"] = 'store_true'
        elif arg_dict["default"] == True:
            arg_dict["action"] = 'store_false'

    # Special case for lists args
    elif isinstance(arg_dict["type"], list):
        arg_dict["nargs"] = '*'
        arg_dict["type"] = arg_dict["type"][0]

    parser.add_argument(*arg_names, **arg_dict)


def jhelp(f: "python function or method"):
    """
    Display a Markdown pretty help message for functions and class methods (default __init__ is a class is passed)
    jhelp also display default values and type annotations if available.
    The docstring synthax should follow the same synthax as the one used for this function
    * f
        Function or method to display the help message for
    """
    # Private import as this is only needed if using jupyter
    from IPython.core.display import display, Markdown

    f_doc = doc_func(f)
    arg_doc = make_arg_dict(f)

    # Signature and function documentation
    s = "**{}** ({})\n\n{}\n\n---\n\n".format(f.__name__, ", ".join(arg_doc.keys()), f_doc)

    # Args doc
    for arg_name, arg_val in arg_doc.items():
        # Arg signature section
        s += "* **{}**".format(arg_name)
        if "default" in arg_val:
            if arg_val["default"] == "":
                s += " (default: \"\")".format(arg_val["default"])
            else:
                s += " (default: {})".format(arg_val["default"])
        if "required" in arg_val:
            s += " (required)"
        if "type" in arg_val:
            if isinstance(arg_val["type"], type):
                s += " [{}]".format(arg_val["type"].__name__)
            elif isinstance(arg_val["type"], list):
                s += " [list({})]".format(arg_val["type"][0].__name__)
            else:
                s += " [{}]".format(arg_val["type"])
        s += "\n\n"
        # Arg doc section
        if "help" in arg_val:
            s += "{}\n\n".format(arg_val["help"])

    # Display in Jupyter
    display(Markdown(s))


def head(fp, n=10, sep="\t", max_char_col=50, comment=None):
    """
    Emulate linux head cmd. Handle gziped files and bam files
    * fp
        Path to the file to be parse.
    * n
        Number of lines to print starting from the begining of the file (Default 10)
    """
    line_list = []

    # Get lines
    try:
        open_fun, open_mode = (gzip.open, "rt") if fp.endswith(".gz") else (open, "r")
        with open_fun(fp, open_mode) as fh:
            line_num = 0
            while line_num < n:
                l = next(fh).strip()
                if comment and l.startswith(comment):
                    continue
                if sep:
                    line_list.append(l.split(sep))
                else:
                    line_list.append(l)
                line_num += 1

    except StopIteration:
        pass

    # Add padding if sep given
    if sep:
        try:
            # Find longest elem per col
            col_len_list = [0 for _ in range(len(line_list[0]))]
            for ls in line_list:
                for i in range(len(ls)):
                    len_col = len(ls[i])
                    if len_col > max_char_col:
                        col_len_list[i] = max_char_col
                    elif len_col > col_len_list[i]:
                        col_len_list[i] = len_col

            # Add padding
            line_list_tab = []
            for ls in line_list:
                s = ""
                for i in range(len(ls)):
                    len_col = col_len_list[i]
                    len_cur_col = len(ls[i])
                    if len_cur_col <= len_col:
                        s += ls[i] + " " * (len_col - len_cur_col) + " "
                    else:
                        s += ls[i][0 : len_col - 3] + "..."
                line_list_tab.append(s)
            line_list = line_list_tab

        # Fall back to non tabulated display
        except IndexError:
            return head(fp=fp, n=n, sep=None)

    for l in line_list:
        print(l)
    print()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup

# Long description from README file
with open("README.md", "r") as fh:
    long_description = fh.read()

# Collect info in a dictionary for setup.py
setup(
    name="NanoCount",
    description="NanoCount estimates transcripts abundance from Oxford Nanopore *direct-RNA sequencing* datasets, using an expectation-maximization approach like RSEM, Kallisto, salmon, etc to handle the uncertainty of multi-mapping reads",
    version="0.4.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/a-slide/NanoCount/",
    author="Adrien Leger",
    author_email="aleg@ebi.ac.uk",
    license="MIT",
    python_requires=">=3.6",
    classifiers=["Development Status :: 3 - Alpha", "Intended Audience :: Science/Research", "Topic :: Scientific/Engineering :: Bio-Informatics", "License :: OSI Approved :: MIT License", "Programming Language :: Python :: 3"],
    install_requires=["tqdm>=4.51.0", "numpy>=1.19.4", "pysam>=0.16.0", "pandas>=1.1.4",  "colorlog>=4.1.0"],
    packages=["NanoCount"],
    entry_points={"console_scripts": ["NanoCount=NanoCount.__main__:main"]},
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ~~~~~~~~~~~~~~IMPORTS~~~~~~~~~~~~~~ #
# Standard library imports
from collections import *

# Third party imports
import pysam
import pandas as pd
from tqdm import tqdm

# Local imports
from NanoCount.Read import Read
from NanoCount.common import *

# ~~~~~~~~~~~~~~MAIN FUNCTION~~~~~~~~~~~~~~ #
class NanoCount:

    # ~~~~~~~~~~~~~~MAGIC METHODS~~~~~~~~~~~~~~ #
    def __init__(
        self,
        alignment_file: str,
        count_file: str = "",
        filter_bam_out: str = "",
        min_alignment_length: int = 50,
        keep_suplementary: bool = False,
        min_query_fraction_aligned: float = 0.5,
        sec_scoring_threshold: float = 0.95,
        sec_scoring_value: str = "alignment_score",
        convergence_target: float = 0.005,
        max_em_rounds: int = 100,
        extra_tx_info: bool = False,
        primary_score: str = "alignment_score",
        max_dist_3_prime: int = 50,
        max_dist_5_prime: int = -1,
        verbose: bool = False,
        quiet: bool = False,
    ):
        """
        Estimate abundance of transcripts using an EM
        * alignment_file
            Sorted and indexed BAM or SAM file containing aligned ONT dRNA-Seq reads including secondary alignments
        * count_file
            Output file path where to write estimated counts (TSV format)
        * filter_bam_out
            Optional output file path where to write filtered reads selected by NanoCount to perform quantification estimation (BAM format)
        * min_alignment_length
            Minimal length of the alignment to be considered valid
        * min_query_fraction_aligned
            Minimal fraction of the primary alignment query aligned to consider the read valid
        * sec_scoring_threshold
            Fraction of the alignment score or the alignment length of secondary alignments compared to the primary alignment to be considered valid
            alignments
        * sec_scoring_value
            Value to use for score thresholding of secondary alignments either "alignment_score" or "alignment_length"
        * convergence_target
            Convergence target value of the cummulative difference between abundance values of successive EM round to trigger the end of the EM loop.
        * max_em_rounds
            Maximum number of EM rounds before triggering stop
        * extra_tx_info
            Add transcripts length and zero coverage transcripts to the output file (required valid bam/sam header)
        * primary_score
            Method to pick the best alignment for each read. By default ("alignment_score") uses the best alignment score (AS optional field), but it can be changed to
            use either the primary alignment defined by the aligner ("primary") or the longest alignment ("alignment_length"). choices = [primary, alignment_score, alignment_length]
        * keep_suplementary
            Retain any supplementary alignments and considered them like secondary alignments. Discarded by default.
        * max_dist_3_prime
            Maximum distance of alignment end to 3 prime of transcript. In ONT dRNA-Seq reads are assumed to start from the polyA tail (-1 to deactivate)
        * max_dist_5_prime
            Maximum distance of alignment start to 5 prime of transcript. In conjunction with max_dist_3_prime it can be used to select near full transcript reads
            only (-1 to deactivate).
        * verbose
            Increase verbosity for QC and debugging
        * quiet
            Reduce verbosity
        """

        # Init package
        opt_summary_dict = opt_summary(local_opt=locals())
        self.log = get_logger(name="Nanocount", verbose=verbose, quiet=quiet)

        self.log.warning("Checking options and input files")
        log_dict(opt_summary_dict, self.log.debug, "Options summary")

        # Save args in self variables
        self.alignment_file = alignment_file
        self.count_file = count_file
        self.filter_bam_out = filter_bam_out
        self.min_alignment_length = min_alignment_length
        self.min_query_fraction_aligned = min_query_fraction_aligned
        self.sec_scoring_threshold = sec_scoring_threshold
        self.sec_scoring_value = sec_scoring_value
        self.convergence_target = convergence_target
        self.max_em_rounds = max_em_rounds
        self.extra_tx_info = extra_tx_info
        self.primary_score = primary_score
        self.keep_suplementary = keep_suplementary
        self.max_dist_5_prime = max_dist_5_prime
        self.max_dist_3_prime = max_dist_3_prime

        self.log.warning("Initialise Nanocount")

        # Collect all alignments grouped by read name
        self.log.info("Parse Bam file and filter low quality alignments")
        self.read_dict = self._parse_bam()

        if self.filter_bam_out:
            self.log.info("Write selected alignments to BAM file")
            self._write_bam()

        # Generate compatibility dict grouped by reads
        self.log.info("Generate initial read/transcript compatibility index")
        self.compatibility_dict = self._get_compatibility()

        # EM loop to calculate abundance and update read-transcript compatibility
        self.log.warning("Start EM abundance estimate")

        self.em_round = 0
        self.convergence = 1

        with tqdm(
            unit=" rounds",
            unit_scale=True,
            desc="\tProgress",
            disable=(quiet or verbose),
        ) as pbar:
            # Iterate until convergence threshold or max EM round are reached
            while self.convergence > self.convergence_target and self.em_round < self.max_em_rounds:
                self.em_round += 1
                # Calculate abundance from compatibility assignments
                self.abundance_dict = self._calculate_abundance()
                # Update compatibility assignments
                self.compatibility_dict = self._update_compatibility()
                # Update counter
                pbar.update(1)
                self.log.debug("EM Round: {} / Convergence value: {}".format(self.em_round, self.convergence))

        self.log.info("Exit EM loop after {} rounds".format(self.em_round))
        self.log.info("Convergence value: {}".format(self.convergence))
        if not self.convergence <= self.convergence_target:
            self.log.error("Convergence target ({}) could not be reached after {} rounds".format(self.convergence_target, self.max_em_rounds))

        # Write out results
        self.log.warning("Summarize data")

        self.log.info("Convert results to dataframe")
        self.count_df = pd.DataFrame(self.abundance_dict.most_common(), columns=["transcript_name", "raw"])
        self.count_df.set_index("transcript_name", inplace=True, drop=True)

        self.log.info("Compute estimated counts and TPM")
        self.count_df["est_count"] = self.count_df["raw"] * len(self.read_dict)
        self.count_df["tpm"] = self.count_df["raw"] * 1000000

        # Add extra transcript info is required
        if self.extra_tx_info:
            tx_df = self._get_tx_df()
            self.count_df = pd.merge(self.count_df, tx_df, left_index=True, right_index=True, how="outer")

        # Cleanup and sort
        self.count_df.sort_values(by="raw", ascending=False, inplace=True)
        self.count_df.fillna(value=0, inplace=True)
        self.count_df.index.name = "transcript_name"

        if self.count_file:
            self.log.info("Write file")
            self.count_df.to_csv(self.count_file, sep="\t")

    # ~~~~~~~~~~~~~~PRIVATE METHODS~~~~~~~~~~~~~~ #
    def _parse_bam(self):
        """
        Parse Bam/Sam file, group alignments per reads, filter reads based on
        selection criteria and return a dict of valid read/alignments
        """
        # Parse bam files
        read_dict = defaultdict(Read)
        ref_len_dict = OrderedDict()
        c = Counter()
        with pysam.AlignmentFile(self.alignment_file) as bam:

            # Collect reference lengths in dict
            for name, length in zip(bam.references, bam.lengths):
                ref_len_dict[name] = length

            for idx, alignment in enumerate(bam):
                if alignment.is_unmapped:
                    c["Discarded unmapped alignments"] += 1
                elif alignment.is_reverse:
                    c["Discarded negative strand alignments"] += 1
                elif not self.keep_suplementary and alignment.is_supplementary:
                    c["Discarded supplementary alignments"] += 1
                elif self.min_alignment_length > 0 and alignment.query_alignment_length < self.min_alignment_length:
                    c["Discarded short alignments"] += 1
                elif self.max_dist_3_prime >= 0 and alignment.reference_end <= ref_len_dict[alignment.reference_name] - self.max_dist_3_prime:
                    c["Discarded alignment with invalid 3 prime end"] += 1
                elif self.max_dist_5_prime >= 0 and alignment.reference_start >= self.max_dist_5_prime:
                    c["Discarded alignment with invalid 5 prime end"] += 1
                else:
                    c["Valid alignments"] += 1
                    read_dict[alignment.query_name].add_pysam_alignment(pysam_aligned_segment=alignment, read_idx=idx)

        # Write filtered reads counters
        log_dict(
            d=c,
            logger=self.log.info,
            header="Summary of alignments parsed in input bam file",
        )

        # Filter alignments
        filtered_read_dict = defaultdict(Read)
        c = Counter()

        for query_name, read in read_dict.items():
            # Check if best alignment is valid
            best_alignment = read.get_best_alignment(primary_score=self.primary_score)

            # In case the primary alignment was removed by filters
            if best_alignment:
                if best_alignment.align_score == 0:
                    c["Reads with zero score"] += 1
                elif best_alignment.align_len == 0:
                    c["Reads with zero len"] += 1
                elif best_alignment.query_fraction_aligned < self.min_query_fraction_aligned:
                    c["Reads with low query fraction aligned"] += 1
                else:
                    filtered_read_dict[query_name].add_alignment(best_alignment)
                    c["Reads with valid best alignment"] += 1
                    for alignment in read.get_secondary_alignments_list(primary_score=self.primary_score):

                        # Filter out secondary alignments based on minimap alignment score
                        if self.sec_scoring_value == "alignment_score" and alignment.align_score / best_alignment.align_score < self.sec_scoring_threshold:
                            c["Invalid secondary alignments"] += 1

                        # Filter out secondary alignments based on minimap alignment length
                        elif self.sec_scoring_value == "alignment_length" and alignment.align_len / best_alignment.align_len < self.sec_scoring_threshold:
                            c["Invalid secondary alignments"] += 1

                        # Select valid secondary alignments
                        else:
                            c["Valid secondary alignments"] += 1
                            filtered_read_dict[query_name].add_alignment(alignment)
            else:
                c["Reads without best alignment"] += 1

        if not "Valid secondary alignments" in c:
            self.log.error("No valid secondary alignments found in bam file. Were the reads aligned with minimap `-p 0 -N 10` options ?")

        # Write filtered reads counters
        log_dict(d=c, logger=self.log.info, header="Summary of reads filtered")
        return filtered_read_dict

    def _write_bam(self):
        """"""
        c = Counter()

        # Make list of alignments idx to select
        selected_read_idx = set()
        for read in self.read_dict.values():
            for alignment in read.alignment_list:
                selected_read_idx.add(alignment.read_idx)
                c["Alignments to select"] += 1

        # Select from original bam file and write to output bam file
        with pysam.AlignmentFile(self.alignment_file) as bam_in:
            with pysam.AlignmentFile(self.filter_bam_out, "wb", template=bam_in) as bam_out:
                for read_idx, alignment in enumerate(bam_in):
                    if read_idx in selected_read_idx:
                        bam_out.write(alignment)
                        c["Alignments written"] += 1
                    else:
                        c["Alignments skipped"] += 1

        log_dict(d=c, logger=self.log.info, header="Summary of alignments written to bam")

    def _get_compatibility(self):
        """"""
        compatibility_dict = defaultdict(dict)
        for read_name, read in self.read_dict.items():
            for alignment in read.alignment_list:
                compatibility_dict[read_name][alignment.rname] = score = 1.0 / read.n_alignment

        return compatibility_dict

    def _calculate_abundance(self):
        """
        Calculate the abundance of the transcript set based on read-transcript compatibilities
        """
        abundance_dict = Counter()
        total = 0
        convergence = 0

        for read_name, comp in self.compatibility_dict.items():
            for ref_name, score in comp.items():
                abundance_dict[ref_name] += score
                total += score

        for ref_name in abundance_dict.keys():
            abundance_dict[ref_name] = abundance_dict[ref_name] / total

            if self.em_round > 1:
                convergence += abs(self.abundance_dict[ref_name] - abundance_dict[ref_name])

        if self.em_round == 1:
            self.convergence = 1
        else:
            self.convergence = convergence

        return abundance_dict

    def _update_compatibility(self):
        """
        Update read-transcript compatibility based on transcript abundances
        """
        compatibility_dict = defaultdict(dict)

        for read_name, comp in self.compatibility_dict.items():
            total = 0
            for ref_name in comp.keys():
                total += self.abundance_dict[ref_name]

            for ref_name in comp.keys():
                compatibility_dict[read_name][ref_name] = self.abundance_dict[ref_name] / total

        return compatibility_dict

    def _get_tx_df(self):
        """
        Extract transcript info from bam file header
        """
        try:
            with pysam.AlignmentFile(self.alignment_file) as bam:
                references = bam.references
                lengths = bam.lengths

            return pd.DataFrame(index=references, data=lengths, columns=["transcript_length"])
        # If any error return empty DataFrame silently
        except Exception:
            return pd.DataFrame()
